/*
 * Copyright (c) 2013, Altera Corporation.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  - Redistributions of source code must retain the above copyright notice,
 *  this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

/* Top-level file for the driver.
 * Deal with device init and shutdown, BAR mapping, and interrupts. */

#include "aclpci.h"
#include "acl_init.c"
#include <asm/siginfo.h>    //siginfo
#include <linux/rcupdate.h> //rcu_read_lock
#include <linux/version.h>  //kernel_version
#include <linux/ktime.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
#include <linux/sched/signal.h>
#else
#include <linux/sched.h> // send_sig_info
#endif

//NOTE: https://www.linuxquestions.org/questions/showthread.php?s=32152bb5ccb99819e31c76339f002437&p=6374786#post6374786
//NOTE: https://lore.kernel.org/all/20220225181927.GA368545@bhelgaas/T/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,18,0)
#define DMA_SET_MASK dma_set_mask
#define DMA_SET_COHERENT_MASK dma_set_coherent_mask
#define DEVICE &dev->dev
#else
#define DMA_SET_MASK pci_set_dma_mask
#define DMA_SET_COHERENT_MASK pci_set_consistent_dma_mask
#define DEVICE dev
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR  ("KAYA Instruments Ltd.");
MODULE_DESCRIPTION ("Driver for KAYA Instruments Predator Frame Grabber");
#ifdef MODULE_SUPPORTED_DEVICE
MODULE_SUPPORTED_DEVICE ("KAYA Instruments Predator");
#endif

/* Use Message Signalled Interrupt (MSI).
 * If not used will get many visibly-distinct interrupts for a single
 * logical one (because it takes a while to reset the interrupt in the FPGA).
 * MSIs are faster. HOWEVER, currently seem to loose MSIs once in a while. :( */
#define USE_MSI 1

/* Static function declarations */
//@@@mbl function names, no __exit k_remove
static int  k_probe(struct pci_dev *dev, const struct pci_device_id *id);
static void k_remove(struct pci_dev *dev);

static int __init init_chrdev (struct aclpci_dev *aclpci);

static int __init scan_bars(struct aclpci_dev *aclpci, struct pci_dev *dev);
static int __init map_bars(struct aclpci_dev *aclpci, struct pci_dev *dev);
static void free_bars(struct aclpci_dev *aclpci, struct pci_dev *dev);

//#define CREATE_CLASS_IN_INIT
#ifdef CREATE_CLASS_IN_INIT
static struct class *devclass;
static dev_t chrdev;
#endif

/* Populating kernel-defined data structures */
// static // static is removed to make struct pci_device_id aclpci_ids[] available in acl_init.c via extern
struct pci_device_id aclpci_ids[] = {
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_CHAMELEON_DEVICE_ID) },
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_PREDATOR_DEVICE_ID) },
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_KOMODO_DEVICE_ID) },
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_KOMODO_4R4T_DEVICE_ID) },
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_KOMODO_FIBER_DEVICE_ID) },
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_COMMON_DEVICE_ID) },
  
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_KOMODO_FIBER_CLHS_DEVICE_ID) },
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_KOMODO_FIBER_GIGE_DEVICE_ID) },
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_FNC_DEVICE_ID) },				
 
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_CHAMELEON_II_DEVICE_ID) },   
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_PREDATOR_II_DEVICE_ID) },    
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_KOMODO_II_DEVICE_ID) },      
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_KOMODO_II_4R4T_DEVICE_ID) },
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_KOMODO_II_FIBER_DEVICE_ID) },
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_KOMODO_FIBER_CLHS_II_DEVICE_ID) },
  { PCI_DEVICE(ACL_PCI_ALTERA_VENDOR_ID, ACL_PCI_KOMODO_FIBER_GIGE_II_DEVICE_ID) },
  
  { 0 },
};
MODULE_DEVICE_TABLE(pci, aclpci_ids);

static struct pci_driver aclpci_driver = {
  .name = DRIVER_NAME,
  .id_table = aclpci_ids,
  .probe = k_probe,
  .remove = k_remove,
  /* resume, suspend are optional */
};


struct file_operations aclpci_fileops = {
  .owner =    THIS_MODULE,
  .read =     aclpci_read,
  .write =    aclpci_write,
/*  .ioctl =    aclpci_ioctl, */
  .open =     aclpci_open,
  .release =  aclpci_close,
/*  .llseek = aclpci_llseek, */
};

static int deviceNumber = 0;
/* Allocate /dev/BOARD_NAME device */
static int __init init_chrdev (struct aclpci_dev *aclpci)
{

  int dev_minor =   0;
  int dev_major =   0; 
  int devno = -1;
  int result;
  char deviceName[64];
  int i;

#ifdef CREATE_CLASS_IN_INIT
  aclpci->cdev_num=chrdev;
#else
  /* request major number for device */
//@@@mbl no increment of deviceNumber here, do it later 
  sprintf(deviceName, "%s%d", BOARD_NAME, deviceNumber);
  
//@@@mbl debug output
  ACL_DEBUG (KERN_DEBUG "aclpci deviceName %s",deviceName);
  
  result = alloc_chrdev_region(&aclpci->cdev_num, dev_minor, MAXBOARDS /* one device*/, deviceName);
  if (result < 0)
  {
    ACL_DEBUG (KERN_WARNING "can't get major ID %d", MAJOR(aclpci->cdev_num));
    goto fail_alloc;
  }
#endif
  dev_major = MAJOR(aclpci->cdev_num);
  ACL_DEBUG (KERN_WARNING "major ID %d", dev_major);
  
  devno = MKDEV(dev_major, dev_minor);
    
  cdev_init (&aclpci->cdev, &aclpci_fileops);
  aclpci->cdev.owner = THIS_MODULE;
  aclpci->cdev.ops = &aclpci_fileops;
  result = cdev_add (&aclpci->cdev, devno, MAXBOARDS);
  /* Fail gracefully if need be */
  if (result)
  {
    ACL_DEBUG(KERN_NOTICE "Error %d adding aclpci (%d, %d)", result, dev_major, dev_minor);
    goto fail_add;
  }
//old  ACL_DEBUG (KERN_DEBUG "aclpci = %d:%d", MAJOR(devno), MINOR(devno));
  ACL_DEBUG (KERN_DEBUG "aclpci(%p): %d:%d, current %p", aclpci, MAJOR(devno), MINOR(devno), current);
  for(i = 0; i < ACL_PCI_MAX_PROCESSES; i++)
  {
      ACL_DEBUG (KERN_DEBUG "aclpci->user_task_for_irq[%d]: %p", i, aclpci->user_task_for_irq[i] );
  }

  
//@@@mbl want automatic generation with /dev/predatorDev_%d  
  aclpci->my_class = class_create(THIS_MODULE, "predator");
  if (IS_ERR(aclpci->my_class))
  {
    printk(KERN_NOTICE "Can't create class\n");
    goto fail_add;
  }
  ACL_DEBUG (KERN_DEBUG "class_create predator done");
  
  aclpci->device = device_create (aclpci->my_class, NULL, aclpci->cdev_num, "%s",deviceName);
  if (IS_ERR(aclpci->device))
  {
    printk(KERN_NOTICE "Can't create device\n");
    goto fail_dev_create;
  }
  ACL_DEBUG (KERN_DEBUG "device_create %s done",deviceName);
  deviceNumber++;
  
  return 0;


fail_dev_create:
  class_unregister(aclpci->my_class);
  class_destroy(aclpci->my_class);

  
/* ERROR HANDLING */
fail_add:
  /* free the dynamically allocated character device node */
  unregister_chrdev_region(devno, MAXBOARDS/*count*/);
  
fail_alloc:
  return -1;
}

/*
static int debug_interrupt_counter = 0;
static int debug_interrupt_send_counter = 0;
static int debug_interrupt_skip_counter = 0;
static unsigned long last_time_signal_sent = 0;
*/
void get_wd_debug_info(WD_DEBUG_INFO *pWD_DEBUG_INFO){}
/*{
    pWD_DEBUG_INFO->int1 = debug_interrupt_counter;
    pWD_DEBUG_INFO->int2 = debug_interrupt_send_counter;
    pWD_DEBUG_INFO->int3 = debug_interrupt_skip_counter;
}

unsigned long get_current_time()
{
    struct timeval time;
    do_gettimeofday(&time);
    unsigned long local_time = 1000 * time.tv_sec + time.tv_usec / 1000;

//    start = ktime_get();
//    //unsigned long local_time  = ktime_to_ns(ktime_sub(end, start));
//    unsigned long local_time = ktime_to_ms(start);


    return local_time;
}
*/


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
irqreturn_t aclpci_irq (int irq, void *dev_id, struct pt_regs * not_used)
#else
irqreturn_t aclpci_irq (int irq, void *dev_id)
#endif
{
  bool do_log = false;
  int i;
  static int do_send_send_sig_info = 1;

  struct aclpci_dev *aclpci = (struct aclpci_dev *)dev_id;
  irqreturn_t res;

  // Read ISR status register to identify false interrupt
  if (!readl(aclpci->bar[2] + (unsigned long)0x40) &&
      !readl(aclpci->bar[2] + (unsigned long)0x140)
     )
  {
      return IRQ_NONE;
  }

  if (aclpci == NULL)
  {
      if (do_log)
      ACL_DEBUG (KERN_WARNING "aclpci_irq exiting for 'aclpci == NULL)'");
      return IRQ_NONE;
  }
  
  /* From this point on, this is our interrupt. So return IRQ_HANDLED
   * no matter what (since nobody else in the system will handle this
   * interrupt for us). */
  aclpci->num_handled_interrupts++;



#if !POLLING
/* Send SIGNAL to user program to notify about the kernel update interrupt. */
/* TODO consider https://man7.org/linux/man-pages/man7/netlink.7.html, https://linux.die.net/man/7/netlink*/
  for(i = 0; i < ACL_PCI_MAX_PROCESSES; i++)
  {
    if (aclpci->user_task_for_irq[i] != NULL)
    {
      int ret = 0;
      if (do_send_send_sig_info)
      {
        ret = send_sig_info(SIG_INT_NOTIFY, &aclpci->signal_info, aclpci->user_task_for_irq[i]);
        if (ret < 0)
        {
              /* Can get to this state if the host is suspended for whatever reason.
               * Just print a warning message the first few times. The FPGA will keep
               * the interrupt level high until the kernel done bit is cleared (by the host).
               * See Case:84460. */


          do_send_send_sig_info = 0;
                //if (do_log)
          ACL_DEBUG (KERN_DEBUG "Wasn't able to execute send_sig_info(), will not try again\n");

          aclpci->num_undelivered_signals[i]++;
          if (aclpci->num_undelivered_signals[i] < 5)
          {
                //if (do_log)
                //ACL_DEBUG (KERN_DEBUG "Error sending signal to host! irq_status is 0x%x\n", irq_status);
            ACL_DEBUG (KERN_DEBUG "Error sending signal to host! \n");
          }
          else 
          {
                //if (do_log_irq)
            ACL_DEBUG (KERN_DEBUG "Wasn't able to execute send_sig_info(), ret %d, Removing task from user_task_for_irq[ACL_PCI_MAX_PROCESSES]\n", ret);
                release_irq (aclpci->pci_dev, aclpci, 0);
          }
        }
        else
        {
          if (do_log)
            ACL_DEBUG (KERN_DEBUG "send_sig_info was sent\n");
        }
      }
      else
      {
        if (do_log)
          ACL_DEBUG (KERN_DEBUG "Didn't try to execute send_sig_info()\n");
      }

    }
    else
    {
      if (do_log)
        ACL_DEBUG (KERN_DEBUG "aclpci->user_task is NULL\n");
    }
  }
#else
    ACL_VERBOSE_DEBUG (KERN_WARNING "Kernel update interrupt. Letting host POLL for it.");
#endif
  res = IRQ_HANDLED;

  return res;
}


void load_signal_info (struct aclpci_dev *aclpci)
{

  /* Setup siginfo struct to send signal to user process. Doing it once here
   * so don't waste time inside the interrupt handler. */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,20,0)
  struct kernel_siginfo *info = &aclpci->signal_info;
  memset(info, 0, sizeof(struct kernel_siginfo));
#else
  struct siginfo *info = &aclpci->signal_info;
  memset(info, 0, sizeof(struct siginfo));
#endif
  
  info->si_signo = SIG_INT_NOTIFY;
  /* this is bit of a trickery: SI_QUEUE is normally used by sigqueue from user
   * space,  and kernel space should use SI_KERNEL. But if SI_KERNEL is used the
   * real_time data is not delivered to the user space signal handler function. */
  info->si_code = SI_QUEUE;
  info->si_int = -1;  /* Signal payload. Will be filled later with
                        ACLPCI_CMD_SET_SIGNAL_PAYLOAD cmd from user. */
}


int init_irq (struct pci_dev *dev, void *dev_id)
{
  u32 irq_type;

  struct aclpci_dev *aclpci = (struct aclpci_dev*)dev_id;
  int i, j, rc;

  if (dev == NULL || aclpci == NULL)
  {
    ACL_DEBUG (KERN_WARNING "Invalid inputs to init_irq (%p, %p)", dev, dev_id);
    return -1;
  }
  ACL_DEBUG (KERN_WARNING "VALID inputs to init_irq (%p, %p)", dev, dev_id);
  
  // hold the semaphore and enter critical section but let function be interruptable
  irq_type = IRQF_SHARED; /* IRQF_SHARED   -- allow sharing IRQs with other devices */

    /* Do NOT use PCI_INTERRUPT_LINE config register. Its value is different
     * from dev->irq and doesn't work! Why? Who knows! */

    for(i = 0; i < ACL_PCI_MAX_PROCESSES; i++)
    {
        //ACL_DEBUG (KERN_DEBUG "aclpci->user_task_for_irq[%d]: %p", i, aclpci->user_task_for_irq[i] );
        if (aclpci->user_task_for_irq[i] != NULL)
        {
            continue;
        }
        aclpci->user_task_for_irq[i] = current;
        aclpci->num_undelivered_signals[i] = 0;
        break;
    }
    
    aclpci->irq_requested = 0;
    for(j = 0; j < ACL_PCI_MAX_PROCESSES; j++)
    {
        ACL_DEBUG (KERN_DEBUG "aclpci->user_task_for_irq[%d]: %p", j, aclpci->user_task_for_irq[j] );
        if (aclpci->user_task_for_irq[j] != 0)
        {
            aclpci->irq_requested += 1;
        }
    }
    
    if(i >= ACL_PCI_MAX_PROCESSES)
    {
        ACL_DEBUG (KERN_WARNING "Couldn't not find vacant place for new 'current' to init_irq (%p, %p)", dev, dev_id);
        return -1;
    }

    ACL_DEBUG (KERN_WARNING "IRQ requested %d times", aclpci->irq_requested);
    if(aclpci->irq_requested > 1)
    {
        ACL_DEBUG (KERN_WARNING "Will not repeat request_irq() because aclpci->irq_requested == %d > 1", aclpci->irq_requested);
        return 0;
    }
  
  /* Message Signalled Interrupts. */  
  #if USE_MSI
    if(pci_enable_msi(dev) != 0)
    {
      ACL_DEBUG (KERN_WARNING "Could not enable MSI");
    }
    else
    {
        ACL_DEBUG (KERN_WARNING "Enable MSI succeeded");
        irq_type = 0;  /* No need to share MSI interrupts since they don't use dedicated wires.*/
    }
  #endif

  rc = request_irq (dev->irq, aclpci_irq, irq_type, DRIVER_NAME, dev_id);
  if (rc)
  {
    ACL_DEBUG (KERN_WARNING "Could not request IRQ #%d, error %d", dev->irq, rc);
    return -1;
  }
  ACL_DEBUG (KERN_WARNING "Could YES request IRQ #%d, error %d", dev->irq, rc);

  aclpci->irq_requested += 1;
  if (aclpci->irq_requested > 1)
  {
      ACL_DEBUG (KERN_WARNING "IRQ requested %d times", aclpci->irq_requested);
  }

  aclpci->num_handled_interrupts = 0;
  
  load_signal_info (aclpci);

  return 0;
}


int release_irq (struct pci_dev *dev, void *dev_id, struct task_struct *user_task)
{
  int i, j;
  
  /* Disable interrupts before going away. If something bad happened in
   * user space and the user program crashes, the interrupt assigned to the device
   * will be freed (on automatic close()) call but the device will continue 
   * generating interrupts. Soon the kernel will notice, complain, and bring down
   * the whole system. */
  //mask_irq(aclpci);
    struct aclpci_dev *aclpci = (struct aclpci_dev*)dev_id;
    
    if(!user_task)
    {
        ACL_DEBUG("(!user_task), will use 'current': %p", current);
        user_task = current;
    }
    ACL_DEBUG("looking for user_task %p...", user_task);
    for(i = 0; i < ACL_PCI_MAX_PROCESSES; i++)
    {
        if (aclpci->user_task_for_irq[i] == user_task)
        {
            ACL_DEBUG("found user_task %p at index %d...", aclpci->user_task_for_irq[i], i);
            aclpci->user_task_for_irq[i] = 0;
            break;
        }
    }

    aclpci->irq_requested = 0;
    for(j = 0; j < ACL_PCI_MAX_PROCESSES; j++)
    {
        ACL_DEBUG (KERN_DEBUG "aclpci->user_task_for_irq[%d]: %p", j, aclpci->user_task_for_irq[j] );
        if (aclpci->user_task_for_irq[j] != 0)
        {
            aclpci->irq_requested += 1;
        }
    }

    if(i >= ACL_PCI_MAX_PROCESSES)
    {
        ACL_DEBUG (KERN_WARNING "Couldn't not find 'current' in the user_task_for_irq[ACL_PCI_MAX_PROCESSES] (%p, %p)", dev, dev_id);
        return 0; //this may happen when a process is killed
    }

    if (  aclpci->irq_requested > 0  )
    {
        ACL_DEBUG (KERN_WARNING "NOT Freeing IRQ %d, because 'irq_requested' left: %d", dev->irq, aclpci->irq_requested);
        return 0;
    }

    ACL_DEBUG (KERN_WARNING "Freeing IRQ %d, irq_requested left: %d", dev->irq, aclpci->irq_requested);

    free_irq (dev->irq, aclpci);
    
    ACL_VERBOSE_DEBUG (KERN_DEBUG "Handled %d interrupts", aclpci->num_handled_interrupts);

//@@@changed mbl to num_undelivered_signals[i]
    if (aclpci->num_undelivered_signals[i] > 0)
    {
      ACL_DEBUG (KERN_DEBUG "Number undelivered signals is %d", aclpci->num_undelivered_signals[i]);
    }

/*old
  if (  ((struct aclpci_dev*)aclpci)->irq_requested > 0  )
  {
      ((struct aclpci_dev*)aclpci)->irq_requested -= 1;
      ACL_DEBUG (KERN_WARNING "Freeing IRQ %d, count %d", dev->irq, ((struct aclpci_dev*)aclpci)->irq_requested);
  }
  else
  {
      ACL_DEBUG (KERN_WARNING "NOT Freeing IRQ %d, because it is already freed", dev->irq);
      return 0;
  }

  
  ACL_VERBOSE_DEBUG (KERN_DEBUG "Handled %d interrupts", 
        ((struct aclpci_dev*)aclpci)->num_handled_interrupts);
        
  num_usignals = ((struct aclpci_dev*)aclpci)->num_undelivered_signals;
  if (num_usignals > 0)
  {
    ACL_DEBUG (KERN_DEBUG "Number undelivered signals is %d", num_usignals);
  }
*/    
  #if USE_MSI
    pci_disable_msi (dev);
  #endif
  return 0;
}

/* Find upstream PCIe root node. 
 * Used for re-training and disabling AER. */
static struct pci_dev* find_upstream_dev (struct pci_dev *dev) {
  struct pci_bus *bus = 0;
  struct pci_dev *bridge = 0;
  struct pci_dev *cur = 0;
  int found_dev = 0;
  
  bus = dev->bus;
  if (bus == 0) {
    ACL_DEBUG (KERN_WARNING "Device doesn't have an associated bus!\n");
    return 0;
  }
  
  bridge = bus->self;
  if (bridge == 0) {
    ACL_DEBUG (KERN_WARNING "Can't get the bridge for the bus!\n");
    return 0;
  }
  
  ACL_DEBUG (KERN_DEBUG "Upstream device %x/%x, bus:slot.func %02x:%02x.%02x", 
             bridge->vendor, bridge->device,
             bridge->bus->number, PCI_SLOT(bridge->devfn), PCI_FUNC(bridge->devfn));
             
  ACL_DEBUG (KERN_DEBUG "List of downstream devices:");
  list_for_each_entry (cur, &bus->devices, bus_list) {
    if (cur != 0) {
      ACL_DEBUG (KERN_DEBUG "  %x/%x", cur->vendor, cur->device);
      if (cur == dev) {
        found_dev = 1;
      }
    }
  }
  
  if (found_dev) {
    return bridge;
  } else {
    ACL_DEBUG (KERN_WARNING "Couldn't find upstream device!");
    return 0;
  }
}

static int k_probe(struct pci_dev *dev, const struct pci_device_id *id)
{

  struct aclpci_dev *aclpci = 0;
  int res;
  int i;

  ACL_DEBUG (KERN_DEBUG " probe (dev = 0x%p, pciid = 0x%p)", dev, id);
  ACL_DEBUG (KERN_DEBUG " vendor = 0x%x, device = 0x%x, class = 0x%x, bus:slot.func = %02x:%02x.%02x",
        dev->vendor, dev->device, dev->class, 
        dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));

  acl_set_board_id_based_on_device_id (dev->device);
  if (board_id == -1)
  {
    ACL_DEBUG (KERN_WARNING "\nNO MAPPING FROM DEVICE ID to board_id!\n");
  }
  
  /* Load all board-related constants */
  //acl_init_board_parameters();

  ACL_DEBUG (KERN_DEBUG "board_id is %d",board_id);
  printk(KERN_INFO "%s: board_id is %d", DRIVER_NAME,board_id);
      
  aclpci = kzalloc(sizeof(struct aclpci_dev), GFP_KERNEL);
  if (!aclpci)
  {
    ACL_DEBUG(KERN_WARNING "Couldn't allocate memory!\n");
    goto fail_kzalloc;
  }
  
  sema_init (&aclpci->sem, 1);
  aclpci->pci_dev = dev;
  dev_set_drvdata(&dev->dev, (void*)aclpci);
//old  aclpci->user_pid = -1;
  for(i = 0; i < ACL_PCI_MAX_PROCESSES; i++)
  {
    aclpci->user_task_for_irq[i] = 0;
    aclpci->pFile[i] = 0;
    aclpci->user_task_for_file[i] = 0;
  }
  aclpci->cvp_in_progress = 0;
  aclpci->pci_gen = 0;
  aclpci->pci_num_lanes = 0;
  aclpci->upstream = find_upstream_dev (dev);
  aclpci->irq_requested = 0;
  
  get_pcie_params (aclpci);
        
  aclpci->buffer = kmalloc (BUF_SIZE * sizeof(char), GFP_KERNEL);
  if (!aclpci->buffer)
  {
    ACL_DEBUG(KERN_WARNING "Couldn't allocate memory for buffer!\n");
    goto fail_kmalloc;
  }
  
  res = init_chrdev (aclpci);
  if (res)
  {
    ACL_DEBUG(KERN_WARNING "Couldn't init_chrdev!\n");
    goto fail_chrdev_init;
  }
  
  if (pci_enable_device(dev))
  {
    ACL_DEBUG (KERN_WARNING "pci_enable_device() failed");
    goto fail_enable;
  }

  pci_set_master(dev);

  if (pci_request_regions(dev, DRIVER_NAME))
  {
    goto fail_regions;
  }


  if (!DMA_SET_MASK(DEVICE, DMA_BIT_MASK(64)))
  {
    DMA_SET_COHERENT_MASK(DEVICE, DMA_BIT_MASK(64));
    /* use 64-bit DMA */
    ACL_DEBUG(KERN_DEBUG "Using a 64-bit DMA mask.\n");
  }
  else if (!DMA_SET_MASK(DEVICE, DMA_BIT_MASK(32)))
  {
    ACL_DEBUG(KERN_DEBUG "Could not set 64-bit DMA mask.\n");
    DMA_SET_COHERENT_MASK(DEVICE, DMA_BIT_MASK(32));
        /* use 32-bit DMA */
    ACL_DEBUG(KERN_DEBUG "Using a 32-bit DMA mask.\n");
  }
  else
  {
    ACL_DEBUG(KERN_DEBUG "No suitable DMA possible.\n");
          /** @todo Choose proper error return code */
    return -1;
  }

// map BARs 
  scan_bars(aclpci, dev);  
  if (map_bars(aclpci, dev))
  {
    goto fail_map_bars;
  }

  return 0;


/* ERROR HANDLING */
fail_map_bars:
  pci_release_regions(dev);
  pci_disable_device (dev);
  
fail_regions:

fail_enable:
  unregister_chrdev_region (aclpci->cdev_num, MAXBOARDS);
  
fail_chrdev_init:
  kfree (aclpci->buffer);
  
fail_kmalloc:
  kfree (aclpci);
  
fail_kzalloc:
  return -1;
}



static int __init scan_bars(struct aclpci_dev * UNUSED(aclpci), struct pci_dev *dev)
{
  int i;
  for (i = 0; i < ACL_PCI_NUM_BARS; i++) {
    unsigned long bar_start = pci_resource_start(dev, i);
    if (bar_start) {
      unsigned long VARIABLE_IS_NOT_USED bar_end = pci_resource_end(dev, i);
      unsigned long VARIABLE_IS_NOT_USED bar_flags = pci_resource_flags(dev, i);
      ACL_DEBUG (KERN_DEBUG "BAR[%d] 0x%08lx-0x%08lx flags 0x%08lx",
         i, bar_start, bar_end, bar_flags);
    }
  }
  return 0;
}


/**
 * Map the device memory regions into kernel virtual address space
 * after verifying their sizes respect the minimum sizes needed, given
 * by the bar_min_len[] array.
 */
static int __init map_bars(struct aclpci_dev *aclpci, struct pci_dev *dev)
{
  int i;
  for (i = 0; i < ACL_PCI_NUM_BARS; i++){
    unsigned long bar_start = pci_resource_start(dev, i);
    unsigned long bar_end = pci_resource_end(dev, i);
    unsigned long bar_length = bar_end - bar_start + 1;
    aclpci->bar_length[i] = bar_length;

    if (!bar_start || !bar_end) {
      aclpci->bar_length[i] = 0;
      continue;
    }

    if (bar_length < 1) {
      ACL_DEBUG (KERN_WARNING "BAR #%d length is less than 1 byte", i);
      continue;
    }

    /* map the device memory or IO region into kernel virtual
     * address space */  
    aclpci->bar[i] = ioremap (bar_start, bar_length);

    if (!aclpci->bar[i]) {
      ACL_DEBUG (KERN_WARNING "Could not map BAR #%d.", i);
      return -1;
    }

    ACL_DEBUG (KERN_DEBUG "BAR[%d] mapped at 0x%p with length %lu.", i,
         aclpci->bar[i], bar_length);
  }
  return 0;
}  



static void free_bars(struct aclpci_dev *aclpci, struct pci_dev *dev) {

  int i;
  for (i = 0; i < ACL_PCI_NUM_BARS; i++) {
    if (aclpci->bar[i]) {
      pci_iounmap(dev, aclpci->bar[i]);
      aclpci->bar[i] = NULL;
    }
  }
}


static void k_remove(struct pci_dev *dev)
{

  struct aclpci_dev *aclpci = 0;
  ACL_DEBUG (KERN_DEBUG ": dev is %p", dev);
  
  if (dev == 0)
  {
    ACL_DEBUG (KERN_WARNING ": dev is 0");
    return;
  }
  
  aclpci = (struct aclpci_dev*) dev_get_drvdata(&dev->dev);
  if (aclpci == 0)
  {
    ACL_DEBUG (KERN_WARNING ": aclpci_dev is 0");
    return;
  }

  ACL_DEBUG (KERN_DEBUG ": device destroy %d", aclpci->cdev_num);
  device_destroy(aclpci->my_class,aclpci->cdev_num);
  class_destroy(aclpci->my_class);
  
  cdev_del (&aclpci->cdev);
  unregister_chrdev_region (aclpci->cdev_num, MAXBOARDS);
  free_bars (aclpci, dev);
  pci_disable_device(dev);
  pci_release_regions(dev);

  kfree (aclpci->buffer);
  kfree (aclpci);
}


/* Initialize the driver module (but not any device) and register
 * the module with the kernel PCI subsystem. */
static int __init aclpci_init(void)
{
  int init_ret = 0;
#ifdef CREATE_CLASS_IN_INIT
  int rc;
#endif

  ACL_DEBUG (KERN_DEBUG "----------------------------");
  ACL_DEBUG (KERN_DEBUG "Driver version: %s", ACL_DRIVER_VERSION);
  
  printk(KERN_INFO "%s: Version %s %s\n", DRIVER_NAME, ACL_DRIVER_VERSION, DRIVER_VENDOR);


  /* Following XIMEA sample we are performing class creation here - removed from probe*/
#ifdef CREATE_CLASS_IN_INIT
  rc = alloc_chrdev_region(&chrdev, 0, 1, BOARD_NAME);
  if(rc < 0)
  {
      ACL_DEBUG(KERN_DEBUG "Char dev allocation failed\n");
      printk(KERN_ERR "%s: failed to register chrdev\n",DRIVER_NAME);

      return rc;
  }
  devclass = class_create(THIS_MODULE, DRIVER_NAME);
  if(IS_ERR(devclass))
  {
      ACL_DEBUG(KERN_DEBUG "Class creation failed\n");
      rc = PTR_ERR(devclass);
      unregister_chrdev_region(chrdev, MAXBOARDS);
      return rc;
  }
#endif
  // End of class creation

  /* register this driver with the PCI bus driver */
  init_ret = pci_register_driver(&aclpci_driver);
  ACL_DEBUG (KERN_DEBUG "pci_register_driver ret: %d", init_ret);
  return init_ret;
}

static void __exit aclpci_exit(void)
{
  ACL_DEBUG (KERN_DEBUG "aclpci_exit...");

  /* unregister this driver from the PCI bus driver */
  pci_unregister_driver(&aclpci_driver);
}


module_init (aclpci_init);
module_exit (aclpci_exit);
