/*
 * Copyright (c) 2013, Altera Corporation.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  - Redistributions of source code must retain the above copyright notice,
 *  this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

/* DMA logic imlementation.
 *
 * The basic flow of DMA transfer is as follows:
 *  1. Pin user memory (a contiguous set of address in processor address space)
 *     to get a list of physical pages (almost never contiguous list of 4KB
 *     blocks).
 *  2. Setup ATT (address translation table) entries for physical page address 
 *     to "row #" mapping in the FPGA PCIe core (a "row number" is just a
 *     counter that is passed in DMA descriptor as a host address).
 *  3. Create and send DMA descriptor (src addr, dest addr, and lengh) to the DMA
 *     core to perform the transfer. 
 *  4. Go to step 2 if have not transfered all currently pinned memory yet.
 *  5. Go to step 1 if need to pin more memory.
 *
 * DMA controller sends an interrupt whenever work for a single DMA descriptor
 * is complete. The driver also checks "done_count" to see how many descriptors
 * completed. This is in case the driver misses an interrupt.
 *
 * To keep interrupt logic simple (i.e. no work-queues), the user has to stall
 * using the following logic until DMA is done:
 *
 *   while (!dma->is_idle())
 *      dma->update(0);
 *
 * The DMA logic is complicated because there are a number of limits of what can
 * be done in one shot. Here they are (all are constants in hw_pcie_constants.h):
 *  - How much user memory can be pinned at one time.
 *  - Size of ATT table in hardware
 *  - Number of outstanding descriptors the hardware can have.
 *
 * Due to hardware restrictions, can only do DMA for 32-byte aligned start
 * addresses (on both host and device) AND 32-byte aligned lengths.
 * Also, need to use a separate descriptor for transfers that do NOT start or
 * end on page boundary. DMA engine does NOT do DMA for these cases, so these
 * transfers are very slow. */



#include <linux/mm.h>
#include <linux/scatterlist.h>
#include <linux/sched.h>
#include <asm/page.h>
#include <linux/spinlock.h>
#include <linux/jiffies.h>
#include "aclpci.h"

#if (KAYA_CUDA_BUILD > 0)
#if (KAYA_CUDA_BUILD == KAYA_CUDA_DGPU)
#pragma message("Building with NV_BUILD_DGPU")
#include <nv-p2p.h>
#else
#pragma message("Building without NV_BUILD_DGPU")
#endif
#endif // #if (KAYA_CUDA_BUILD > 0)

#define MAP_UNMAP_PAGES 1

//NOTE: https://docs.kernel.org/core-api/dma-api.html
//NOTE: https://github.com/torvalds/linux/blob/master/include/linux/dma-direction.h
//NOTE: https://github.com/torvalds/linux/blob/master/include/linux/dma-mapping.h
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,18,0)
#define DMA_MAP_PAGE dma_map_page
#define DMA_UNMAP_PAGE dma_unmap_page
#define DMA_SYNC_SINGLE_FOR_DEVICE dma_sync_single_for_device
#define DMA_SYNC_SINGLE_FOR_CPU dma_sync_single_for_cpu
#define ACL_PCI_DEVICE &aclpci->pci_dev->dev
#else
#define DMA_BIDIRECTIONAL PCI_DMA_BIDIRECTIONAL
#define DMA_FROM_DEVICE PCI_DMA_FROMDEVICE
#define DMA_TO_DEVICE PCI_DMA_TODEVICE
#define DMA_MAP_PAGE pci_map_page
#define DMA_UNMAP_PAGE pci_unmap_page
#define DMA_SYNC_SINGLE_FOR_DEVICE pci_dma_sync_single_for_device
#define DMA_SYNC_SINGLE_FOR_CPU pci_dma_sync_single_for_cpu
#define ACL_PCI_DEVICE aclpci->pci_dev
#endif

__inline int KERNEL_DIRECTION_FROM_USER_DIRECTION(int dmaOptions)
{
    int direction = DMA_BIDIRECTIONAL;

    if(dmaOptions & WD_DMA_FROM_DEVICE)
    {
        if(!(dmaOptions & WD_DMA_TO_DEVICE))
        {
            direction = DMA_FROM_DEVICE;
        }
    }
    else
    {
        if(dmaOptions & WD_DMA_TO_DEVICE)
        {
            direction = DMA_TO_DEVICE;
        }
    }
    return direction;
}

int lock_dma_buffer (struct aclpci_dev *aclpci, WD_DMA *dma)
{

  int ret;
  unsigned int i = 0;
  //struct aclpci_dma *d = &(aclpci->dma_data);
  //ssize_t start_page, end_page, num_pages;
  //u64 ej, startj = get_jiffies_64();

  /*TODO remove debug output*/
  //ACL_DEBUG (KERN_DEBUG "enter lock_dma_buffer()");
  /***************************/
  /* num_pages that [addr, addr+len] map to. */
 
/*
  ACL_DEBUG (KERN_DEBUG "addr is: %p", dma->pUserAddr);

  start_page = (ssize_t)dma->pUserAddr >> PAGE_SHIFT;
//  ACL_DEBUG (KERN_DEBUG "start_page is: %ld, size is %d", start_page, dma->dwBytes);

  end_page = ((ssize_t)dma->pUserAddr + dma->dwBytes - 1) >> PAGE_SHIFT;
//  ACL_DEBUG (KERN_DEBUG "end_page is: %ld", end_page);

  num_pages = end_page - start_page + 1;
  ACL_DEBUG (KERN_DEBUG "num_pages is: %ld", num_pages);
*/

  ACL_DEBUG (KERN_DEBUG "addr is: %p 0x%lx size is %d pages %d",dma->pUserAddr,(unsigned long)dma->pUserAddr,dma->dwBytes,dma->dwPages);
  
  dma->Page = (WD_DMA_PAGE*)kzalloc ( sizeof(WD_DMA_PAGE) * dma->dwPages, GFP_KERNEL );
  if (dma->Page == NULL)
  {
    ACL_DEBUG (KERN_DEBUG "Couldn't allocate array of %u page descriptors!", dma->dwPages);
    return -EFAULT;
  }

  dma->ppInternal = (void**)kzalloc ( sizeof(struct page*) * dma->dwPages, GFP_KERNEL );
  if (dma->ppInternal == NULL)
  {
    ACL_DEBUG (KERN_DEBUG "Couldn't allocate array of %u 'page' pointers!", dma->dwPages);
    return -EFAULT;
  }

  dma->dma_addrs = (dma_addr_t*)kzalloc ( sizeof(dma_addr_t) * dma->dwPages, GFP_KERNEL );
  if (dma->dma_addrs == NULL)
  {
    ACL_DEBUG (KERN_DEBUG "Couldn't allocate array of %u dma_addr_t's!", dma->dwPages);
    return -EFAULT;
  }
  //ACL_VERBOSE_DEBUG (KERN_DEBUG "pages = [%p, %p), dma_addrs = [%p, %p)",
    //                 dma->pages, dma->pages+dma->dwPages, dma->dma_addrs, dma->dma_addrs+dma->dwPages);

//  ACL_DEBUG (KERN_DEBUG "Allocations: Page %p ppInt %p addrs %p",dma->Page,dma->ppInternal,dma->dma_addrs);
    
  /* pin user memory and get set of physical pages back in 'p' ptr. */
  ret = aclpci_get_user_pages((unsigned long)dma->pUserAddr & PAGE_MASK, (size_t)dma->dwPages, (struct page**)dma->ppInternal);
  if (ret != 0)
  {
    ACL_DEBUG (KERN_DEBUG "Couldn't pin all user pages %d free kernel allocated memory", ret);
    kfree(dma->dma_addrs);
    kfree(dma->ppInternal);
    kfree(dma->Page);
    dma->ppInternal = NULL;
    dma->dma_addrs = NULL;
    dma->Page= NULL;
    return -EFAULT;
  }


/*
  dma->first_page_offset = (unsigned long)dma->pUserAddr & ACL_PCIE_DMA_ATT_PAGE_ADDR_MASK;
  dma->last_page_offset = (unsigned long)(dma->pUserAddr + dma->dwBytes) & ACL_PCIE_DMA_ATT_PAGE_ADDR_MASK;

  for (i = 0; i < dma->dwPages; i++)
  {
    struct page *cur = dma->ppInternal[i];
    //ACL_DEBUG (KERN_DEBUG "p[%d] = 0x%p", i, cur);
    if (cur != NULL)
    {
       //ACL_DEBUG (KERN_DEBUG "  phys_addr = 0x%llx", page_to_phys(cur));
    dma->Page[i].dwBytes = PAGE_SIZE;
    if(0 == i) // calculate the start of first page
    {
           dma->Page[i].dwBytes -= dma->first_page_offset;
    }

    if(((dma->dwPages - 1) == i) && (dma->last_page_offset > 0)) // calculate end of last page
    {
           dma->Page[dma->dwPages - 1].dwBytes -= (PAGE_SIZE - dma->last_page_offset);
    }
        dma->Page[i].pPhysicalAddr = pci_map_page (aclpci->pci_dev, cur, 0, dma->Page[i].dwBytes, KERNEL_DIRECTION_FROM_USER_DIRECTION(dma->dwOptions));
    }
  }
*/

  dma->first_page_offset = (unsigned long)dma->pUserAddr & ACL_PCIE_DMA_ATT_PAGE_ADDR_MASK;
  dma->last_page_offset = (unsigned long)(dma->pUserAddr + dma->dwBytes) & ACL_PCIE_DMA_ATT_PAGE_ADDR_MASK;

  /* map pages for PCI access. */
  #if MAP_UNMAP_PAGES
  for (i = 0; i < dma->dwPages; i++)
  {
    struct page *cur = dma->ppInternal[i];
    //ACL_DEBUG (KERN_DEBUG "p[%d] = 0x%p", i, cur);
    if (cur != NULL) {
      //ACL_DEBUG (KERN_DEBUG "  phys_addr = 0x%llx", page_to_phys(cur));

//old dma_addr_t phys = pci_map_page (aclpci->pci_dev, cur, 0, PAGE_SIZE, KERNEL_DIRECTION_FROM_USER_DIRECTION(dma->dwOptions));
      dma_addr_t phys = DMA_MAP_PAGE(ACL_PCI_DEVICE, cur, 0, PAGE_SIZE, KERNEL_DIRECTION_FROM_USER_DIRECTION(dma->dwOptions));
      if (phys == 0)
      {
        ACL_DEBUG (KERN_DEBUG "  Couldn't pci_map_page %d",i);
        kfree(dma->dma_addrs);
        kfree(dma->ppInternal);
        kfree(dma->Page);
        dma->ppInternal = NULL;
        dma->dma_addrs = NULL;
        dma->Page= NULL;
        return -EFAULT;
      }
      dma->Page[i].pPhysicalAddr = dma->dma_addrs[i] = phys;
      dma->Page[i].dwBytes = PAGE_SIZE;
      if(0 == i) // calculate the start of first page
      {
          dma->Page[i].dwBytes -= dma->first_page_offset;
          dma->Page[i].pPhysicalAddr += dma->first_page_offset;
      }

      if(((dma->dwPages - 1) == i) && (dma->last_page_offset > 0)) // calculate end of last page
      {
          dma->Page[dma->dwPages - 1].dwBytes -= (PAGE_SIZE - dma->last_page_offset);
      }

    }
  }
  #endif

  //ej = get_jiffies_64();

  //d->m_pin_time += (ej - startj);
 // d->m_lock_time += (ej - startj);
  return 0;
}


void unlock_dma_buffer (struct aclpci_dev *aclpci, WD_DMA *dma)
{
  int i;

#if VERBOSE_DEBUG
  u64 ej, startj = get_jiffies_64();
#endif

  ACL_DEBUG (KERN_DEBUG "addr is: %p 0x%lx size is %d pages %d",dma->pUserAddr,(unsigned long)dma->pUserAddr,dma->dwBytes,dma->dwPages);
  
  #if MAP_UNMAP_PAGES
  /* Unmap pages to make the data available for CPU */      
  for (i = 0; i < dma->dwPages; i++)
  {
    struct page *cur = dma->ppInternal[i];
    // ACL_DEBUG (KERN_DEBUG "p[%d] = %p", i, cur);
    if (cur != NULL) {
      dma_addr_t phys = dma->dma_addrs[i];
      DMA_UNMAP_PAGE(ACL_PCI_DEVICE, phys, PAGE_SIZE, KERNEL_DIRECTION_FROM_USER_DIRECTION(dma->dwOptions));
    }
  }
  #endif

  /* Unpin pages */
  aclpci_release_user_pages ((struct page**)dma->ppInternal, dma->dwPages);

//  ACL_DEBUG (KERN_DEBUG "Free ppInt %p addrs %p",dma->ppInternal,dma->dma_addrs);

  /* TODO: try to re-use these buffers on future allocs */
  kfree (dma->ppInternal);
  kfree (dma->dma_addrs);

  //ej = get_jiffies_64();
  ACL_VERBOSE_DEBUG (KERN_DEBUG  "DMA: Unpinned ? pages in %u usec",
                          //dma->num_pages,
                          jiffies_to_usecs(ej - startj));
                          
  dma->ppInternal = NULL;
  dma->dma_addrs = NULL;
  //d->m_pin_time += (ej - startj);
  //d->m_unlock_time += (ej - startj);
}

#if (KAYA_CUDA_BUILD > 0)
#if (KAYA_CUDA_BUILD == KAYA_CUDA_DGPU)
#define GPU_PAGE_SHIFT	16
#else
#define GPU_PAGE_SHIFT	12
#endif
#define GPU_PAGE_SIZE	(((u64)1) << GPU_PAGE_SHIFT)
#define GPU_PAGE_OFFSET	(GPU_PAGE_SIZE - 1)
#define GPU_PAGE_MASK	(~GPU_PAGE_OFFSET)

struct pevb_cuda_surface {
    struct aclpci_dev *aclpci; //struct pevb_file		*pevb_file;
    u64				va;
    u64				offset;
    u64				len;
    int				handle;
    struct nvidia_p2p_page_table	*page_table;
};

static void cuda_p2p_free_callback(void *data)
{
    struct pevb_cuda_surface *cusurf = data;
}

int cuda_lock_dma_buffer (struct aclpci_dev *aclpci, WD_DMA *dma)
{
    u64 aligned_len;
    int ret;

    struct pevb_cuda_surface *cusurf;
    cusurf = kzalloc(sizeof(*cusurf), GFP_KERNEL);
    if (!cusurf)
        return -ENOMEM;

    cusurf->aclpci = aclpci;
    cusurf->va = ((u64)dma->pUserAddr) & GPU_PAGE_MASK;
    cusurf->offset = ((u64)dma->pUserAddr) & GPU_PAGE_OFFSET;
    cusurf->len = dma->dwBytes;
    aligned_len = (cusurf->offset + cusurf->len + GPU_PAGE_SIZE - 1) & GPU_PAGE_MASK;


    ret = nvidia_p2p_get_pages(
#if (KAYA_CUDA_BUILD == KAYA_CUDA_DGPU)
        0, 0,
#endif
        cusurf->va, aligned_len, &cusurf->page_table,
        cuda_p2p_free_callback, cusurf);
    if (ret < 0) {
        kfree(cusurf);
        return ret;
    }

    //mutex_lock(&pevb_file->lock);
    //cusurf->handle = idr_alloc(&pevb_file->cuda_surfaces, cusurf, 0, 0,
    //    GFP_KERNEL);
    //mutex_unlock(&pevb_file->lock);

}

void cuda_unlock_dma_buffer (struct aclpci_dev *aclpci, WD_DMA *dma)
{

}
#endif // #if (KAYA_CUDA_BUILD > 0)

void sync_cache_for_memory(struct aclpci_dev *aclpci, WD_DMA *dma, bool forDevice)
{
    int i;
    for (i = 0; i < dma->dwPages; i++)
    {
      struct page *cur = dma->ppInternal[i];
      if (cur != NULL)
      {
        dma_addr_t phys = dma->dma_addrs[i];
        if(forDevice)
        {
            DMA_SYNC_SINGLE_FOR_DEVICE(ACL_PCI_DEVICE, phys, PAGE_SIZE, KERNEL_DIRECTION_FROM_USER_DIRECTION(dma->dwOptions));
        }
        else
        {
            DMA_SYNC_SINGLE_FOR_CPU(ACL_PCI_DEVICE, phys, PAGE_SIZE, KERNEL_DIRECTION_FROM_USER_DIRECTION(dma->dwOptions));
        }
      }
    }
}
/*
void sync_memory_for_cpu (struct aclpci_dev *aclpci, WD_DMA *dma)
{
    int i;
    for (i = 0; i < dma->dwPages; i++)
    {
      struct page *cur = dma->ppInternal[i];
      if (cur != NULL)
      {
        dma_addr_t phys = dma->dma_addrs[i];

      }
    }
}
*/
