/*
 * Copyright (c) 2013, Altera Corporation.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  - Redistributions of source code must retain the above copyright notice,
 *  this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */


/* Implementation of all I/O functions except DMA transfers.
 * See aclpci_dma.c for DMA code.
 */

#include <linux/jiffies.h>
#include <linux/sched.h>
#include "aclpci.h"


static ssize_t aclpci_rw_large (void *dev_addr, void __user* user_addr, ssize_t len, char *buffer, int reading);

/* Given (bar_id, device_addr) pair, make sure they're valid and return
 * the resulting address. errno will contain error code, if any. */
void* aclpci_get_checked_addr (int bar_id, void *device_addr, size_t count,
                               struct aclpci_dev *aclpci, ssize_t *errno,
                               int print_error_msg)
{

  if (bar_id >= ACL_PCI_NUM_BARS)
  {
    ACL_DEBUG (KERN_WARNING "Requested read/write from BAR #%d. Only have %d BARs!",
               bar_id, ACL_PCI_NUM_BARS);
    *errno = -EFAULT;
    return 0;
  }
  /* Make sure the final address is within range */
  if (((unsigned long)device_addr + count) > aclpci->bar_length[bar_id])
  {
    if (print_error_msg)
    {
      ACL_DEBUG (KERN_WARNING "Requested read/write from BAR #%d from range (%lu, %lu). Length is %lu. BAR length is only %lu!",
                 bar_id, 
                 (unsigned long)device_addr,
                 (unsigned long)device_addr + count, 
                 count,
                 aclpci->bar_length[bar_id]);
    }
    *errno = -EFAULT;
    return 0;
  }

  *errno = 0;
  return (void*)(aclpci->bar[bar_id] + (unsigned long)device_addr);  
}

#define INIT_RELEASE_IRQ 1

/* Response to user's open() call */
int aclpci_open(struct inode *inode, struct file *pFile)
{
  struct aclpci_dev *aclpci = 0;
  int result = 0;
  int i;

  /* pointer to containing data structure of the character device inode */
  aclpci = container_of(inode->i_cdev, struct aclpci_dev, cdev);


    for(i = 0; i < ACL_PCI_MAX_PROCESSES; i++)
    {
        if (aclpci->pFile[i] != NULL)
        {
            continue;
        }
        aclpci->pFile[i] = pFile;
        aclpci->user_task_for_file[i] = current;
        ACL_DEBUG("Registered user_task %p for file %p", aclpci->user_task_for_file[i], pFile);
        break;
    }
    if(i >= ACL_PCI_MAX_PROCESSES)
    {
        ACL_DEBUG (KERN_WARNING "Couldn't not find vacant place for new 'user_task_for_file' to aclpci_open()");
        return -1;
    }

    /* create a reference to our device state in the opened file */
    pFile->private_data = aclpci;
    ACL_DEBUG (KERN_DEBUG "aclpci = %p, current %p: pid = %d (%s), pFile %p", aclpci, current, current->pid, current->comm, pFile);

//old  aclpci->user_pid = current->pid;
//old  aclpci->user_task = current;
  
  aclpci->global_mem_segment = 0;

  result = 0;
  
//done:
  return result;
}


/* Response to user's close() call. Will also be called by the kernel
 * if the user process dies for any reason. */
int aclpci_close(struct inode *inode, struct file *pFile)
{
  int i;
  ssize_t result = 0;
  struct aclpci_dev *aclpci = (struct aclpci_dev *)pFile->private_data;
  struct task_struct *user_task_for_file = 0;

  ACL_DEBUG (KERN_DEBUG "aclpci = %p, current %p: pid = %d (%s), pFile %p", aclpci, current, current->pid, current->comm, pFile);

  for(i = 0; i < ACL_PCI_MAX_PROCESSES; i++)
  {
    if (aclpci->pFile[i] != pFile)
    {
      continue;
    }
    user_task_for_file = aclpci->user_task_for_file[i];
    ACL_DEBUG("FOUND user_task %p, pid = %d (%s) for file %p", user_task_for_file, current->pid, current->comm, pFile);
    aclpci->pFile[i] = 0;
    aclpci->user_task_for_file[i] = 0;
    break;
  }
  if(i >= ACL_PCI_MAX_PROCESSES)
  {
    ACL_DEBUG("Couldn't not find 'user_task_for_file': %p", pFile);
  }

  
#if INIT_RELEASE_IRQ
  release_irq (aclpci->pci_dev, aclpci,user_task_for_file);
#endif

  return result;
}


/* Read a small number of bytes and put them into user space */
ssize_t aclpci_read_small (void *read_addr, void __user* dest_addr, ssize_t len)
{
  ssize_t copy_res = 0;
  switch (len)
  {
      case 1:
      {
        u8 d = readb ( read_addr );
        copy_res = copy_to_user ( dest_addr, &d, sizeof(d) );
        break;
      }
      case 2:
      {
        u16 d = readw ( read_addr );
        copy_res = copy_to_user ( dest_addr, &d, sizeof(d) );
        break;
      }
      case 4:
      {
        u32 d = readl ( read_addr );
        copy_res = copy_to_user ( dest_addr, &d, sizeof(d) );
        break;
      }
      case 8:
      {
        u64 d = readq ( read_addr );
        copy_res = copy_to_user ( dest_addr, &d, sizeof(d) );
        break;
      }
      default:
        break;
  }

  if (copy_res)
  {
    return -EFAULT;
  }
  else
  {
    return 0;
  }
}


/* Write a small number of bytes taken from user space */
ssize_t aclpci_write_small (void *write_addr, void __user* src_addr, ssize_t len)
{

  ssize_t copy_res = 0;
  switch (len)
  {
      case 1:
      {
        u8 d;
        copy_res = copy_from_user ( &d, src_addr, sizeof(d) );
        writeb ( d, write_addr );
        break;
      }
      case 2:
      {
        u16 d;
        copy_res = copy_from_user ( &d, src_addr, sizeof(d) );
        writew ( d, write_addr );
        break;
      }
      case 4:
      {
        u32 d;
        copy_res = copy_from_user ( &d, src_addr, sizeof(d) );
        writel ( d, write_addr );
        break;
      }
      case 8:
      {
        u64 d;
        copy_res = copy_from_user ( &d, src_addr, sizeof(d) );
        writeq ( d, write_addr );
        break;
      }
      default:
        break;
  }

  if (copy_res)
  {
    return -EFAULT;
  }
  else
  {
    return 0;
  }
}



/* Read or Write arbitrary length sequency starting at read_addr and put it into
 * user space at dest_addr. if 'reading' is set to 1, doing the read. If 0, doing
 * the write. */
static ssize_t aclpci_rw_large (void *dev_addr, void __user* user_addr, ssize_t len, char *buffer, int reading)
{
  size_t bytes_left = len;
  size_t i, num_missed;
  u32 *ibuffer = (u32*)buffer;
  char *cbuffer;
  size_t offset, num_to_read;
  size_t chunk = BUF_SIZE;
#if VERBOSE_DEBUG
  u64 startj, ej;
#endif
  u64 sj = 0, acc_readj = 0, acc_transfj = 0;

#if VERBOSE_DEBUG
  startj = get_jiffies_64();
#endif
  if(reading) 
   ACL_DEBUG("aclpci_rw_large read %ld bytes",len);
  else
   ACL_DEBUG("aclpci_rw_large write %ld bytes",len);

  
  /* Reading upto BUF_SIZE values, one int at a time, and then transfer
   * the buffer at once to user space. Repeat as necessary. */
  while (bytes_left > 0)
  {
    if (bytes_left < BUF_SIZE)
    {
      chunk = bytes_left;
    }
    else
    {
      chunk = BUF_SIZE;
    }

    if (!reading)
    {
      sj = get_jiffies_64();
      if (copy_from_user (ibuffer, user_addr, chunk))
      {
        return -EFAULT;
      }
      acc_transfj += get_jiffies_64() - sj;
    }

    /* Read one u32 at a time until fill the buffer. Then copy the whole
     * buffer at once to user space. */
    sj = get_jiffies_64();
    num_to_read = chunk / sizeof(u32);
    for (i = 0; i < num_to_read; i++)
    {
      if (reading)
      {
        ibuffer[i] = readl ( ((u32*)dev_addr) + i);
      }
      else
      {
        writel ( ibuffer[i], ((u32*)dev_addr) + i );
      }
    }

    /* If length is not a multiple of sizeof(u32), will miss last few bytes.
     * In that case, read it one byte at a time. This can only happen on
     * last iteration of the while() loop. */
    offset = num_to_read * sizeof(u32);
    num_missed = chunk - offset;
    cbuffer = (char*)(ibuffer + num_to_read);

    for (i = 0; i < num_missed; i++)
    {
      if (reading)
      {
        cbuffer[i] = readb ( (u8*)(dev_addr) + offset + i );
      }
      else
      {
        writeb ( cbuffer[i], (u8*)(dev_addr) + offset + i );
      }
    }
    acc_readj += get_jiffies_64() - sj;

    if (reading)
    {
      sj = get_jiffies_64();
      if (copy_to_user (user_addr, ibuffer, chunk))
      {
        return -EFAULT;
      }
      acc_transfj += get_jiffies_64() - sj;
    }

    dev_addr += chunk;
    user_addr += chunk;
    bytes_left -= chunk;
  }

#if VERBOSE_DEBUG
  ej = get_jiffies_64();
#endif
//  ACL_VERBOSE_DEBUG (KERN_DEBUG "Spent %u msec %sing %lu bytes", jiffies_to_msecs(ej - startj), reading ? "read" : "writ", len);

//  ACL_VERBOSE_DEBUG (KERN_DEBUG "  Dev access %u msec. User space transfer %u msec",
//                        jiffies_to_msecs(acc_readj),
//                        jiffies_to_msecs(acc_transfj));
  return 0;
}

/* Set CRA window so raw_user_ptr is "visible" to the BAR.
 * Return pointer to use to access the user memory */
//static void* aclpci_set_segment (struct aclpci_dev *aclpci, void * raw_user_ptr) {

//  ssize_t cur_segment = ((ssize_t)raw_user_ptr) / ACL_PCIE_MEMWINDOW_SIZE;
//  aclpci_set_segment_by_val (aclpci, cur_segment);

//  /* Can use the return value in all read/write functions in this file now */
//  return (void*)((ssize_t)raw_user_ptr % ACL_PCIE_MEMWINDOW_SIZE);
//}


/* Both start and end, user and device addresses must be 
 * 64-byte aligned to use DMA */
//int aligned_request (struct aclpci_cmd *cmd, size_t count) {
  
//  return (( (unsigned long)cmd->user_addr   & DMA_ALIGNMENT_BYTE_MASK) |
//          ( (unsigned long)cmd->device_addr & DMA_ALIGNMENT_BYTE_MASK) |
//          ( count                           & DMA_ALIGNMENT_BYTE_MASK)
//         ) == 0;
//}

ssize_t aclpci_rw(struct file *file,
                  char __user *buf,
                  size_t count, loff_t *pos,
                  int reading);


/* Response to user's read() call */
ssize_t aclpci_read(struct file *file,
                    char __user *buf,
                    size_t count,
                    loff_t *pos)
{

  return aclpci_rw (file, buf, 0, pos, 1 /* reading */);
}


/* Response to user's write() call */
ssize_t aclpci_write(struct file *file, const char __user *buf, size_t count, loff_t *pos)
{
  return aclpci_rw (file, (char __user *)buf, 0, pos, 0 /* writing */);
}


/* High-level read/write dispatcher. */
ssize_t aclpci_rw(struct file *file, char __user *buf, size_t count, loff_t *pos, int reading)
{
  struct aclpci_dev *aclpci = (struct aclpci_dev *)file->private_data;
  struct aclpci_cmd __user *ucmd;
  struct aclpci_cmd kcmd;
  //u64 old_segment = 0;
  //int restore_segment = 0;
  void *addr = 0;
  //int aligned = 0;
  int use_dma = 0;
  ssize_t result = 0;
  ssize_t errno = 0;  

//  ACL_DEBUG("Entry");
  
  if (down_interruptible(&aclpci->sem))
  {
    ACL_DEBUG("Exit sem");
    return -ERESTARTSYS;
  }
  
  ucmd = (struct aclpci_cmd __user *) buf;
  if (copy_from_user (&kcmd, ucmd, sizeof(*ucmd)))
  {
    ACL_DEBUG("Exit copy");
    result = -EFAULT;
    goto done;
  }
  
  if (kcmd.bar_id == ACLPCI_CMD_BAR)
  {
    /* This is not a read but a special command. */
    result = aclpci_exec_cmd (aclpci, kcmd, kcmd.buff_size);
    goto done;
  }

//  ACL_DEBUG("NOT In ACLPCI_CMD_BAR");

  /* Only using DMA for large aligned reads/writes on global memory
   * (due to some assumptions inside the DMA code). */
  //aligned = aligned_request (&kcmd, kcmd.buff_size);
  //use_dma = USE_DMA && (kcmd.buff_size >= 1024) &&
   //         aligned && kcmd.bar_id == ACL_PCI_GLOBAL_MEM_BAR;

//  ACL_VERBOSE_DEBUG (KERN_DEBUG " kcmd = {%u, %p, %p}, count = %lu", kcmd.bar_id, (void*)kcmd.device_addr, (void*)kcmd.user_addr, kcmd.buff_size);
  if (!use_dma)
  {
//    ACL_DEBUG("call rw aclpci_get_checked_addr %p",kcmd.device_addr);
    addr = aclpci_get_checked_addr (kcmd.bar_id, kcmd.device_addr, kcmd.buff_size, aclpci, &errno, 1);
    /* If not using DMA, need 'addr', which is mem-mapped virtual address. If that's out of
     * range of the BAR, can't do the operation. For DMA, mem-mapped virtual addresses are not
     * used. So don't care if they're in the BAR range or not. */
    if (errno != 0)
    {
        ACL_DEBUG("aclpci_get_checked_addr failed exit errno %ld",errno);
        result = -EFAULT;
        goto done;
    }
  }

  /* Offset value is always an address offset, not element offset. */
  /* ACL_DEBUG (KERN_DEBUG "Read address is %p", addr); */
  
  switch (kcmd.buff_size)
  {
      case 1:
      case 2:
      case 4:
      //case 8:
      {
        if (reading)
        {
          result = aclpci_read_small (addr, (void __user*) kcmd.user_addr, kcmd.buff_size);
        }
        else
        {
          result = aclpci_write_small (addr, (void __user*) kcmd.user_addr, kcmd.buff_size);
        }
        break;
      }

      default:
      {
        result = aclpci_rw_large (addr, (void __user*) kcmd.user_addr, kcmd.buff_size, aclpci->buffer, reading);
        break;
      }
  }//switch (kcmd.buff_size)
  
done:
  up (&aclpci->sem);

  return result;
}

