/*
 * Copyright (c) 2013, Altera Corporation.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  - Redistributions of source code must retain the above copyright notice,
 *  this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

/* All defines necessary to communicate with the Linux PCIe driver.
 * The actual communication functions are open()/close()/read()/write().
 *
 * Example read call (read single ushort from BAR 0, device address 0x2):
 *   ssize_t f = open ("/dev/de4", O_RDWR);
 *   unsigned short val;
 *   struct aclpci_cmd read_cmd = { 0, ACLPCI_CMD_DEFAULT, 0x2, &val };
 *   read (f, &read_cmd, sizeof(val));
 *
 * See user.c for a tester of all functions and more elaborate examples.
 */

#ifndef PCIE_LINUX_DRIVER_EXPORTS_H
#define PCIE_LINUX_DRIVER_EXPORTS_H


/* if bar_id in aclpci_cmd is set to this, this is a special command,
 * not a usual read/write request. So the command field is used. Otherwise,
 * command field is ignored. */
#define ACLPCI_CMD_BAR 23

/* Values for 'command' field of aclpci_cmd. */

/* Default value -- noop. */
#define ACLPCI_CMD_DEFAULT                0

/* Save/Restore all board PCI control registers to user_addr.
 * Allows user program to reprogram the board without having root
 * priviliges (which is required to change PCI control registers). */
#define ACLPCI_CMD_SAVE_PCI_CONTROL_REGS  1
#define ACLPCI_CMD_LOAD_PCI_CONTROL_REGS  2

/* Lock/Unlock user_addr memory to physical RAM ("pin" it) */
//#define ACLPCI_CMD_PIN_USER_ADDR          3
//#define ACLPCI_CMD_UNPIN_USER_ADDR        4

/* Get m_idle status of DMA */
//#define ACLPCI_CMD_GET_DMA_IDLE_STATUS    5
//#define ACLPCI_CMD_DMA_UPDATE             6

#define ACLPCI_CMD_INTERRUPT_ENABLE       3
#define ACLPCI_CMD_INTERRUPT_DISABLE      4

/* Get vendor_id and device_id of loaded PCIe device */
#define ACLPCI_CMD_GET_DEVICE_ID          7
#define ACLPCI_CMD_GET_VENDOR_ID          8

/* PCIe link status queries (PCIe gen and number of lanes) */
#define ACLPCI_CMD_GET_PCI_GEN            10
#define ACLPCI_CMD_GET_PCI_NUM_LANES      11

/* Set id to receive back on signal from kernel */
#define ACLPCI_CMD_SET_SIGNAL_PAYLOAD     12

/* Get full driver version, as string */
#define ACLPCI_CMD_GET_DRIVER_VERSION     13

//
// KAYA Additional commands:
//
#define ACLPCI_CMD_KAYA_DEBUG             14
#define ACLPCI_CMD_DMA_SGLOCK             15
#define ACLPCI_CMD_DMA_SGUNLOCK           16

/* sync memory cache */
#define ACLPCI_CMD_SYNC_FOR_CPU           17
#define ACLPCI_CMD_SYNC_FOR_DEVICE        18

/* Get vendor_id and device_id of loaded PCIe device */
#define ACLPCI_CMD_GET_BUS          19
#define ACLPCI_CMD_GET_SLOT         20
#define ACLPCI_CMD_GET_FUNC         21

//
// KAYA Additional commands for CUDA operations:
//
#ifndef KAYA_CUDA_BUILD
#define KAYA_CUDA_BUILD 0
#endif
#if (KAYA_CUDA_BUILD > 0)
#define ACLPCI_CMD_PIN_CUDA_IGPU         22
#define ACLPCI_CMD_UNPIN_CUDA_IGPU       23
#define ACLPCI_CMD_PIN_CUDA_DGPU         24
#define ACLPCI_CMD_UNPIN_CUDA_DGPU       25
#endif


#define ACLPCI_CMD_MAX_CMD          26



#define SIG_INT_NOTIFY 44       // Signal from driver to user (hal) to notify about hw interrupt
#define SIG_INT_HANDLER_FLAG 45 // Signal used for indication about being in a signal handler (https://stackoverflow.com/a/7293582/1468415)

/* Main structure to communicate any command (including read/write)
 * from user space to the driver. */
struct aclpci_cmd {

  /* base address register of PCIe device. device_addr is interpreted
   * as an offset from this BAR's start address. */
  unsigned int bar_id;
  
  /* Special command to execute. Only used if bar_id is set
   * to ACLPCI_CMD_BAR. */
  unsigned int command;
  
  /* Address in device space where to read/write data. */
  void* device_addr;
  
  /* Address in user space where to write/read data.
   * Always virtual address. */
  void* user_addr;
  void* user_addr2;
  unsigned int buff_size;
};

#ifndef KYDRIVER_BUILD
#include "ky_defines.h" // TODO: causes compilation error while building driver
#endif

typedef unsigned int DWORD;
typedef unsigned long long KPTR;
typedef uint64_t DMA_ADDR;


typedef struct
{
    DMA_ADDR pPhysicalAddr; // Physical address of page.
    DWORD dwBytes;          // Size of page.
} WD_DMA_PAGE, WD_DMA_PAGE_V80;

enum {
    WD_DMA_KERNEL_BUFFER_ALLOC = 0x1, // The system allocates a contiguous buffer.
        // The uWD_Openser does not need to supply linear address.

    WD_DMA_KBUF_BELOW_16M = 0x2, // If DMA_KERNEL_BUFFER_ALLOC is used,
        // this will make sure it is under 16M.

    WD_DMA_LARGE_BUFFER = 0x4, // If DMA_LARGE_BUFFER is used,
        // the maximum number of pages are dwPages, and not
        // WD_DMA_PAGES. If you lock a user buffer (not a kernel
        // allocated buffer) that is larger than 1MB, then use this
        // option and allocate memory for pages.

    WD_DMA_ALLOW_CACHE = 0x8,  // Allow caching of contiguous memory.

    WD_DMA_KERNEL_ONLY_MAP = 0x10, // Only map to kernel, dont map to user-mode.
        // relevant with DMA_KERNEL_BUFFER_ALLOC flag only

    WD_DMA_FROM_DEVICE = 0x20, // memory pages are locked to be written by device

    WD_DMA_TO_DEVICE = 0x40, // memory pages are locked to be read by device

    WD_DMA_TO_FROM_DEVICE = (WD_DMA_FROM_DEVICE | WD_DMA_TO_DEVICE), // memory pages are
        // locked for both read and write

    WD_DMA_ALLOW_64BIT_ADDRESS = 0x80, // Use this value for devices that support
                                    // 64-bit DMA addressing.

    WD_DMA_ALLOW_NO_HCARD = 0x100, // allow memory lock without hCard
};

typedef struct
{
    DWORD hDma;             // Handle of dma buffer.
    void* pUserAddr;        // Beginning of buffer.
    KPTR  pKernelAddr;      // Kernel mapping of kernel allocated buffer
    DWORD dwBytes;          // Size of buffer.
    DWORD dwOptions;        // Allocation options:
                // DMA_KERNEL_BUFFER_ALLOC, DMA_KBUF_BELOW_16M,
                // DMA_LARGE_BUFFER
                // DMA_ALLOW_CACHE, DMA_KERNEL_ONLY_MAP,
                // DMA_FROM_DEVICE,
                            // DMA_TO_DEVICE, DMA_ALLOW_64BIT_ADDRESS
    DWORD dwPages;          // Number of pages in buffer.
    DWORD hCard;            // Handle of relevant card as received from
                            // WD_CardRegister()
    WD_DMA_PAGE* Page;

    // Former 'dma_t' members (that are not covered above):
    void** ppInternal;        //
    KPTR *dma_addrs;   /* one for each struct page */
    unsigned int first_page_offset;
    unsigned int last_page_offset;
} WD_DMA, WD_DMA_V80;

typedef struct
{
    int int1;
    int int2;
    int int3;
    int int4;
    int int5;
    int int6;
    int int7;
    int int8;
    int int9;
    int int10;
    int int11;
    int int12;
} WD_DEBUG_INFO;

#define MAXBOARDS 16

#endif /* PCIE_LINUX_DRIVER_EXPORTS_H */
